/******************************************************************
 * ap-carray.c
 * - character array module 
 * 11 JUN 00: created by arim@i.kyushu-u.ac.jp
 *            a modification of ap-iarray.c
 * $Id: ap-carray.c,v 1.1.1.1 2004/02/14 04:30:12 arim Exp $
 ******************************************************************/

#include <assert.h>
#include <stdio.h>  /* for BUFSIZ */
#include <stdlib.h>  
#include <math.h>   /* for pow(), and log() */
#include <ctype.h>  /* isgraph(), isspace() */
#include <limits.h>  /* INT_MAX */
#include "ap-util.h"
#include "ap-carray.h"

/*
 * Intarray
 */

//#define DEBUG

/******************************************************************
 * Allocation and free routines
 * 
 ******************************************************************/

Charray ca_alloc(int len)
{
  Charray tmp; /* Note: Charray is a pointer to a struct. */
  char *tmpval;

  if ((tmp = (Charray) malloc(sizeof(struct _charray))) == NULL)
    error_exit("ca_alloc: cannot allocate memory\n");

  len = max(len, DEFAULT_ARRAY_LEN); /* for efficiency with realloc() */
  if ((tmpval = (char *) calloc(len, sizeof(char))) == NULL)
    error_exit("ca_alloc: cannot allocate memory\n");
  tmp->val    = tmpval;
  tmp->len    = 0;      /* A new array is always empty */
  tmp->maxlen = len;
  return tmp;
}

void ca_free(Charray tmp)
{
  if (tmp == NULL) return;
  free(tmp->val);
  /* do nothing for tmp->len */
  /* do nothing for tmp->maxlen */
  free(tmp);
}

void ca_make_null(Charray tmp)
{
  tmp->len = 0; 
  /* do nothing for tmp->val */
  /* do nothing for tmp->maxlen */
}

/******************************************************************
 * Append routines
 * 
 ******************************************************************/

/* Inport ap_expansion_length() from ap-util.h */

void ca_realloc(Charray ca, int size)
{
  int maxlen;
  char *tmpval; 

#ifdef DEBUG
  assert(ca != NULL);
#endif
  /*
   * No need for expansion
   */
  if (ca->maxlen >= size)
    return; /* do nothing */
  
  /*
   * Expanding ca so that it has enough length to keep size cells. 
   */
  maxlen = ap_expansion_length(size);
  /* debug("ca_realloc: ap_expansion_length=%d\n", maxlen); */
  if ((tmpval = (char *) realloc(ca->val, maxlen * sizeof(char))) == NULL)
    error_exit("ca_alloc: cannot reallocate memory\n");
  ca->val    = tmpval;
  ca->maxlen = maxlen; 
  /* do nothing with ca->len */
}

void ca_append(Charray ca, Charray cb)
{ /* - append the contents of ib with the end of ca
   * - ca is reallocated if it has enough memory to have cb. 
   */
  int n, m; 

#ifdef DEBUG
  assert(ca != NULL && cb != NULL);
#endif
  ca_realloc(ca, (ca->len + cb->len));
  for (m = ca->len, n = 0; n < cb->len; m++, n++)
    ca->val[m] = cb->val[n];
  ca->len = m; 
}

void ca_copy(Charray ca, Charray cb)
{ /* - copies the contents of cb into the cells of ca
   * - it is truncated if cb is longer than the allocated part of ca. 
   */
  int n, maxlen;
//#ifdef DEBUG
  assert(ca != NULL && cb != NULL);
//#endif
  ca_realloc(ca, cb->len);
  ca_make_null(ca);  
  for (n = 0; n < cb->len; n++)
    ca_push_back(ca, cb->val[n]);
}

/* inline cannot used for g++ due to undefined refenrece at ld */
//inline 
void ca_push_back(Charray ca, char c)
{ /* - append a character with the end of ca. 
   * - expands the array if it is full. 
   */
  int len;
#ifdef DEBUG
  assert(ca != NULL);
#endif
  len = ca->len; 
  ca_realloc(ca, len + 1); /* ensuring ca->len < len+1 */
  ca->val[len] = c;
  ca->len = len + 1;
}

/* inline cannot used for g++ due to undefined refenrece at ld */
//inline 
void push_back(Charray ca, char c)
{ /* - append a character with the end of ca. 
   * - expands the array if it is full. 
   */
  int len;
#ifdef DEBUG
  assert(ca != NULL);
#endif
  len = ca->len; 
  ca_realloc(ca, len + 1); /* ensuring ca->len < len+1 */
  ca->val[len] = c;
  ca->len = len + 1;
}

/* inline cannot used for g++ due to undefined refenrece at ld */
//inline 
char ca_pop(Charray ca) 
{ /* - append a character with the end of ca. 
   * - expands the array if it is full. 
   */
#ifdef DEBUG
    assert(ca != NULL);
#endif
    return ca->val[--(ca->len)];
}

/******************************************************************
 * Print routines
 * 
 ******************************************************************/

char printable_char(char c) {
  if (isgraph(c))
    return c;
  else if (c == '\0')
    return '#'; 
  else if (isspace(c))
    return '_'; 
  else
    return '@'; 
}

void ca_print(Charray ca)
{
  int n;
#ifdef DEBUG
  assert(ca != NULL);
#endif
  if (ca == NULL) return; 
  for (n = 0; n < ca->len; n++)
    printf("%c", printable_char(ca->val[n]));
}

void ca_report(Charray ca)
{
    int maxlen = DEFAULT_COLWIDTH * 3; 
    if(ca == NULL) {
	printf("\t ia\t == NULL!\n");
	return;
    }
    printf("\t ca\t len=%d\t maxlen=%d\n", ca->len, ca->maxlen);
    if (ca->len <= maxlen) {
	printf("\t "); ca_print(ca); print_nl();
    } else {
	printf("\t skipping the output... \n");
    }
}

/*
 * Print routines for unsigned char as small integers. 
 */

void ca_int_print(Charray ca)
{
  int n;
#ifdef DEBUG
  assert(ca != NULL);
#endif
  if (ca == NULL) return; 
  for (n = 0; n < ca->len; n++)
    printf("%u:", (unsigned int) ca->val[n]); /* printing as an unsigned integer */
}

void ca_int_report(Charray ca)
{
    int maxlen = DEFAULT_COLWIDTH * 3; 
    if(ca == NULL) {
	printf("\t ia\t == NULL!\n");
	return;
    }
    printf("\t ca\t len=%d\t maxlen=%d\n", ca->len, ca->maxlen);
    if (ca->len <= maxlen) {
	printf("\t "); ca_int_print(ca); print_nl();
    } else {
	printf("\t skipping the output... \n");
    }
}

/******************************************************************
 * File I/O routines
 * 
 ******************************************************************/

#ifndef min
#define min(a, b) ((a)<=(b) ? (a) : (b))
#endif
#define PERMS 0666
#define WRITABLE(n, nmax) (min(BUFSIZ, nmax - n))
/* BUFSIZ == 1024 in stdio.h on Solaris2.6 */

int ca_write_disk(Charray ca, char *outname)
{   /* - writes the contents of ca to a binary file on a disk.  
     * - returns the number of bytes written if succeeded and
     * NULL otherwise. */
    FILE *fp;
    int n     = 0, /* the number of bytes written so far*/
        n_int = 0; /* the number of bytes written at once */
    char *buf;	 /* BUFSIZ == 1024 in stdio.h on Solaris2.6 */

    debug("ca_write_disk: opening output file: %s\n", outname);
    /* Carray: Is it right? */
    if ((fp = fopen(outname, "wb")) == NULL) { /* to open with 666 */
	/* Warning: this mode should be binary */
	printf("ca_write_disk: cannot open %s\n", outname);
	return FALSE;
    }

    debug("ca_write_disk: writing a char array...\n");
    buf = ca->val;
    n   = 0;
    while ((n_int = WRITABLE(n, ca->len)) > 0) {
	//debug("ca_write_disk: n=%d, buf-%d, written=%d ints\n", n, buf, n_int);
	if (fwrite(buf, sizeof(char), n_int, fp) != n_int)
	  error_exit("ca_write_disk: write error at %dth letter on %s\n",
		     n, outname);
	buf = buf + n_int;
	n   = n   + n_int;
    }
    //debug("ca_write_disk: closing output file...\n");
    assert(n == ca->len);
    fclose(fp);
    return n; 
}

int ca_read_disk(Charray ca, char *inname)
{ /* - reads the contents of ca from a binary file on a disk. 
   * - returns the number of bytes written if succeeded and
   * NULL otherwise. */
    FILE *fp;
    char *buf; 
    int n   = 0,	/* the number of letters written */
        n_int = 0;	/* the number of letters to be written */
    
    debug("ca_read_disk: opening input file: %s\n", inname);
    if ((fp = fopen(inname, "rb")) == NULL) {
	/* Warning: the mode should be binary to correct reading */
	printf("ca_read_disk: cannot open %s\n", inname);
	return FALSE;
    }
    
#ifdef DEBUG
    assert(ca != NULL);
#endif
    debug("ca_read_disk: reading an int array...\n");
    buf = ca->val;
    n   = 0;
    //while ((n < maxn_int) && (n_int = read(fd, buf, BUFSIZ)) > 0) { 
    while ((n < ca->len)
	   && (n_int = fread(buf, sizeof(int), BUFSIZ, fp)) > 0) { 
	//debug("ca_read_disk: n=%d, buf %d, read=%d ints\n", n, buf, n_int);
	buf = buf + n_int;
	n   = n   + n_int;
    }
    if (ferror(fp)) { /* error of some kind */
	printf("ca_read_disk: read error on file %s\n", inname);
	return FALSE;
    }
    debug("ca_read_disk: closing input file...\n");
    fclose(fp);
    return n; 
}

/*
 * verification: 24 SEP 99 by arim
 */

int ca_veryfy_disk(Charray ca, char *file_name)
{ /* - Verify if the original ia and its copy on a disk  file_name are
     identical */
  Charray cac;
  int n;
  
  cac = ca_alloc(ca->len); /* temporary */
  if (ca_read_disk(cac, file_name) == FALSE)  {
      printf("ca_verify: cannot read the file: %s\n", file_name);
      return FALSE;
  }
  for (n = 0; n < cac->len; n++) {
      //if (1) 
      //  debug("equiv: n=%d: %d!=%d\n", n, ca->val[n], cac->val[n]);
      if (ca->val[n] != cac->val[n]) {
	  printf("ca and cac differ at %d-th cell!\n", n);
	  return FALSE;
      }
  }
  debug("ca_verify: verification suceeded. \n");  
  return TRUE;
}

/******************************************************************
 * Interface to C-string
 ******************************************************************/

/* ca_append_str:
 * - appends a C-string to the tail of ca. 
 */
Charray ca_append_str(Charray ca, char *str)
{
    int n = 0;

    while (*str != '\0' && n < INT_MAX) {
	ca_push_back(ca, *str);
	str++;
	n++; 
    }
    return ca; 
}

/* ca_copy_str:
 * - overwrite a C-string to the tail of ca.
 */

//inline /* inline cannot used for g++ due to undefined refenrece at ld */
Charray ca_copy_str(Charray ca, char *str)
{
    int n = 0;

    ca_make_null(ca);
    return ca_append_str(ca, str);
}

/* ca_copy_str:
 * - copy the contents of ca into a C-string. 
 */
void ca_write_to_str(Charray ca, char *str, int len)
{
  int n;
  assert(ca != NULL);
  len = min(len, ca->len); 
  for (n = 0; n < len; n++)
    str[n] = ca->val[n]; 
}

/******************************************************************
 * End
 * 
 ******************************************************************/

