/******************************************************************
 * ap-iarray.c
 * - integer array module 
 * 20 SEP 99: created by arim@i.kyushu-u.ac.jp.
 * 21 SEP 99: ia_write_disk() and ia_readJ_disk() are added.
 * 22 SEP 99: - Replacing old K&R read and write routines based on fd 
 *   with ANSI fread and fwrite routines based on fp. This is only for
 *   the upper compatibility for g++. 
 * $Id: ap-iarray.c,v 1.2 2004/02/25 20:31:59 arim Exp $
 ******************************************************************/

#include <assert.h>
#include <stdio.h>  /* for BUFSIZ */
#include <stdlib.h>  
#include "ap-util.h" /* for ap_expansion_length */
#include "ap-iarray.h"


//#define DEBUG
#undef DEBUG

/*
 * Intarray
 */

Intarray ia_alloc(int len)
{
  Intarray tmp; /* Note: Intarray is a pointer to a struct. */
  int *tmpval;

  if ((tmp = (Intarray) malloc(sizeof(struct intarray))) == NULL)
    error_exit("ia_alloc: cannot mallocate memory\n");

  len = max(len, DEFAULT_ARRAY_LEN); /* for efficiency with realloc() */
  if ((tmpval = (int *) calloc(len, sizeof(int))) == NULL)
    error_exit("ia_alloc: cannot callocate memory\n");
  tmp->val    = tmpval;
  tmp->len    = 0;      /* A new array is always empty */
  tmp->maxlen = len;
  tmp->minelem = 0;
  return tmp;
}

void ia_free(Intarray temp)
{
  if (temp == NULL) return;
  free(temp->val);
  /* do nothing for tmp->len */
  /* do nothing for tmp->maxlen */
  free(temp);
}

void ia_make_null(Intarray tmp)
{
    assert(tmp != NULL);
    tmp->len = 0; 
    /* do nothing for tmp->val */
    /* do nothing for tmp->maxlen */
}

/* Inport ap_expansion_length() from ap-util.h */

void ia_realloc(Intarray ia, int size)
{ /* - expands the maxlen of Intarray ia if maxlen > size.
   * /
  /* Note: Here, we cannot use debug() routines because this ia_realloc()
   * itself used in the definition of debug() routines */
  int maxlen;
  int *tmpval;
  assert(ia != NULL); 
  if (ia->maxlen >= size)
    return; /* do nothing */

#ifdef DEBUG
  printf("ia_realloc: called with ia=%d, size=%d\n", ia, size);
  fflush(stdout);
#endif
  maxlen = ap_expansion_length(size);
#ifdef DEBUG
  printf("ia_realloc: expanding from %d to %d...\n", ia->maxlen, maxlen); fflush(stdout); 
#endif
  if ((tmpval = (int *) realloc(ia->val, maxlen * sizeof(int))) == NULL)
    error_exit("ia_alloc: cannot reallocate memory %d %d\n", maxlen, size);
  ia->val    = tmpval;
  ia->maxlen = maxlen; 
  /* do nothing with ia->len */
#ifdef DEBUG
  assert(ia->maxlen >= size);
#endif
}

void ia_assert_defined(Intarray ia)
{
    assert(ia != NULL);
    assert(ia->val != NULL);
    assert(ia->maxlen >= 0);
    assert(ia->len <= ia->maxlen);
}

//inline /* inline cannot used for g++ due to undefined refenrece at ld */
void ia_push_back(Intarray ia, int x)
{ /* - append a character with the end of ca. 
   * - expands the array if it is full. 
   */
  int len;
#ifdef DEBUG
  assert(ia != NULL);
#endif
  len = ia->len; 
  ia_realloc(ia, len + 1);
  ia->val[len] = x;
  ia->len = len + 1;
}

void ia_append(Intarray ia, Intarray ib)
{ /* - append the contents of ib with the end of ia
   * - ia is reallocated if it has enough memory to have ib. 
   */
  int n, m; 

#ifdef DEBUG
  assert(ia != NULL && ib != NULL); 
#endif
  ia_realloc(ia, (ia->len + ib->len));  /* expand ia when it is too small */

  for (m = ia->len, n = 0; n < ib->len; m++, n++) {
    ia->val[m] = ib->val[n];
  }
  ia->len = m; 
}

void ia_copy(Intarray ia, Intarray ib)
{ /* - copies to ia the contents of ib
   * - optimized version
   * - this is as fast as ia_copy below. Please use ia_copy() */
  int n;
  int *val_a, *val_b, *val_b_tail; 
  
#ifdef DEBUG
  assert(ia != NULL && ib != NULL); 
#endif
  ia_realloc(ia, ib->len);
  val_a = ia->val; 
  val_b = ib->val; 
  val_b_tail = val_b + ib->len; 
  while (val_b < val_b_tail) 
    *(val_a++) = *(val_b++);
  ia->len = ib->len;
}

void ia_copy_org(Intarray ia, Intarray ib)
{ /* - copies to ia the contents of ib
   * - this small program with simple loop is as fast as ia_copy() above!
   */
  int n;

#ifdef DEBUG
  assert(ia != NULL && ib != NULL); 
#endif
  ia_realloc(ia, ib->len);
#ifdef DEBUG
  assert(ia->maxlen >= ib->len);
#endif
  for (n = 0; n < ib->len; n++)
    ia->val[n] = ib->val[n];
  ia->len = n; 
}

void ia_copy_slowest(Intarray ia, Intarray ib)
{ /* - copies to ia the contents of ib */
  int n;

#ifdef DEBUG
  assert(ia != NULL && ib != NULL); 
  assert(ia->maxlen >= ib->len);
#endif
  ia_make_null(ia);
  for (n = 0; n < ib->len; n++)
    ia_push_back(ia, ib->val[n]);
}

void ia_ncopy(Intarray ia, Intarray ib, int len)
{ /* - copies to ia the initial len letters in ib. 
   * - Already tuned as in ia_copy. 
   */
  int n;

#ifdef DEBUG
  assert(ia != NULL && ib != NULL); 
#endif
  ia_realloc(ia, len);
  for (n = 0; n < len; n++)
    ia->val[n] = ib->val[n];
  ia->len = n; 
}



int ia_last_val(Intarray ia)
{ /* - returns the value of the last cell */
    assert(ia->len > 0); 
    return ia->val[ia->len - 1];
}

/*
 * stack
 */

Intstack ia_stack_alloc(void)
{
    Intstack tmp;
    
    tmp = ia_alloc(0);
    ia_make_null(tmp);
    return tmp;
}

void ia_stack_free(Intstack tmp)
{
    ia_free(tmp);
}

void ia_push(Intstack ia, int x)
{ /* - inserts an element at the end of ca. 
   * - expands the array if it is full. 
   */
    assert(ia != NULL);
    ia_push_back(ia, x); 
}

int ia_pop(Intstack ia)
{   /* - retrieve and delete the last element of ia */
    int x;

    assert(ia != NULL);
    assert(ia->len > 0); /* critical */
    x = ia->val[ia->len - 1];
    ia->len--;
}

int ia_top(Intstack ia)
{   /* - returns the last element of ia */
    assert(ia != NULL); 
    assert(ia->len > 0); /* critical */
    return (ia->val[ia->len - 1]);
}

int ia_is_empty(Intstack ia)
{
    assert(ia != NULL);
    return (ia->len == 0);
}

/*
 * print
 */
void ia_print(Intarray ia)
{
  int n;
#ifdef DEBUG
  assert(ia != NULL);
  assert(ia->len <= ia->maxlen);
#endif
  if (ia == NULL) return; 
  for (n = 0; n < ia->len; n++)
    //debug_printf("%d:", ia->val[n]);
    printf("%d:", ia->val[n]);
}

void ia_report(Intarray ia)
{
    int maxlen = DEFAULT_COLWIDTH; 
    if(ia == NULL) {
      printf("\t ia\t == NULL!\n");
      return;
    }
    printf("\t ia\t len=%d\t maxlen=%d\n", ia->len, ia->maxlen);
    if (ia->len <= maxlen) {
      printf("\t "); ia_print(ia); print_nl();
    } else {
      printf("\t skipping the output... \n");
    }
}

void ia_report_colwidth(Intarray ia, int colwidth)
{
    int maxlen = DEFAULT_COLWIDTH / colwidth; 
    if(ia == NULL) {
      printf("\t ia\t == NULL!\n");
      return;
    }
    printf("\t ia\t len=%d\t maxlen=%d\n", ia->len, ia->maxlen);
    if (ia->len <= maxlen) {
      printf("\t "); ia_print(ia); print_nl();
    } else {
      printf("\t skipping the output... \n");
    }
}

#ifdef COMMENT
void ia_report(Intarray ia)
{
    int maxlen = DEFAULT_COLWIDTH * 2; 
    if(ia == NULL) {
      printf("\t ia\t == NULL!\n");
      return;
    }
    debug_printf("\t ia\t len=%d\t maxlen=%d\n", ia->len, ia->maxlen);
    if (ia->len <= maxlen) {
      debug_printf("\t "); ia_print(ia); print_nl();
    } else {
      debug_printf("\t skipping the output... \n");
    }
}
#endif

/******************************************************************
 * Interface to C-int-array
 ******************************************************************/

/* ca_append_str:
 * - appends a C-string to the tail of ca. 
 */
Intarray ia_append_array(Intarray ia, int *array, int len)
{
    int n = 0;

    while (n < len) {
	ia_push_back(ia, array[n]);
	n++; 
    }
    return ia; 
}

/* ia_copy_array:
 * - overwrite a C-int-array to the tail of ia.
 */

//inline /* inline cannot used for g++ due to undefined refenrece at ld */
Intarray ia_copy_array(Intarray ia, int *array, int len)
{
    int n = 0;

    ia_make_null(ia);
    return ia_append_array(ia, array, len);
}

/* ia_copy_array:
 * - copy the contents of ca into a C-int-array. 
 */
void ia_write_to_array(Intarray ia, int *array, int len)
{
  int n;
  assert(ia != NULL);
  len = min(len, ia->len); 
  for (n = 0; n < len; n++)
    array[n] = ia->val[n]; 
}

/******************************************************************
 * Disk Input/Output
 ******************************************************************/

#define PERMS 0666
#define WRITABLE(n, nmax) (min(IOBUFSIZE, nmax - n))
/* IOBUFSIZE == 1024 in stdio.h on Solaris2.6 */

int ia_write_disk(Intarray ia, char *outname)
{   /* - writes the contents of ia to a binary file on a disk. 
     * - returns the number of bytes written if succeeded and
     * NULL otherwise. */
    FILE *fp;
    int n   = 0, /* the number of bytes written so far*/
    n_int = 0;	 /* the number of bytes written at once */
    int *buf;	 /* IOBUFSIZE == 1024 in stdio.h on Solaris2.6 */

    //debug("ia_write_disk: opening output file: %s\n", outname);
    if ((fp = fopen(outname, "wb")) == NULL) { /* to open with 666 */
	/* Warning: this mode should be binary */
	printf("ia_write_disk: cannot open %s\n", outname);
	return FALSE; 
    }

#ifdef DEBUG
    assert(ia != NULL);
#endif
    //debug("ia_write_disk: writing an int array...\n");
    buf = ia->val;
    n   = 0;
    while ((n_int = WRITABLE(n, ia->len)) > 0) {
      //debug("ia_write_disk: n=%d, buf-%d, written=%d ints\n", n, buf, n_int);
	if (fwrite(buf, sizeof(int), n_int, fp) != n_int)
	  error_exit("ia_write_disk: write error at %dth letter on %s\n",
		     n, outname);
	buf = buf + n_int;
	n   = n   + n_int;
    }
    //debug("ia_write_disk: closing output file...\n");
    assert(n == ia->len);
    fclose(fp);
    return n; 
}

int ia_read_disk(Intarray ia, char *inname)
{ /* - reads the contents of ia from a binary file on a disk.  
   * - returns the number of bytes written if succeeded and
   * NULL otherwise. */
    FILE *fp;
    //int *buf; 
    int buf[IOBUFSIZE+1]; 
    int n   = 0,		/* the number of letters written */
    n_int = 0;			/* the number of letters to be written */
#ifdef DEBUG
    debug_push(ON);
#endif
    debug("ia_read_disk: opening input file: %s\n", inname);
    if ((fp = fopen(inname, "rb")) == NULL) {
	/* Warning: the mode should be binary to correct reading */
	debug("ia_read_disk: cannot open %s\n", inname);
	return FALSE; 
    }
    
#ifdef DEBUG
    assert(ia != NULL);
#endif
    debug("ia_read_disk: reading an int array...\n");
    n   = 0;
    debug("ia_read_disk: init: n=%d, ia->len %d, read=%d ints\n",
	  n, ia->len, n_int);
    while ((n_int = fread(buf, sizeof(int), IOBUFSIZE, fp)) > 0) { 
	//debug("ia_read_disk: n=%d, buf %d, read=%d ints\n", n, buf, n_int);
	ia_append_array(ia, buf, n_int);
	n   = n   + n_int;
    }
    if (ferror(fp)) {	/* error of some kind */
	printf("ia_read_disk: read error on file %s\n", inname);
	return FALSE;
    }
    debug("ia_read_disk: letter read %d\n", n);
    debug("ia_read_disk: closing input file...\n");
    fclose(fp);
#ifdef DEBUG
    debug_pop();
#endif
    return n; 
}

/*
 * verification: 24 SEP 99 by arim
 */

int ia_veryfy_disk(Intarray ia, char *file_name)
{ /* - Verify if the original ia and its copy on a disk  file_name are
   * identical.
   * - returns TRUE if identical and FALSE otherwise. */
  Intarray iac;
  int n;
  
#ifdef DEBUG
    assert(ia != NULL);
#endif
  iac = ia_alloc(ia->len); /* temporary */
  if (ia_read_disk(iac, file_name) == FALSE)  {
      printf("ia_verify: cannot read the file: %s\n", file_name);
      return FALSE;
  }
  debug("ia_verify: the length of the original and the copy: %d, %d\n",
	ia->len, iac->len);
  for (n = 0; n < iac->len; n++) {
      //if (1) 
      //  debug("equiv: n=%d: %d!=%d\n", n, ia->val[n], iac->val[n]);
      if (ia->val[n] != iac->val[n]) {
	  printf("ia and iac differ at %d-th cell!\n", n);
	  return FALSE; 
      }
  }
  debug("ia_verify: verification suceeded. \n");  
  return TRUE;
}

/*
 * Array operations for counting
 * - 01 SEP 00
 */

/* Code optimization */
void ia_map(Intarray ia, Intarray ib)
{   /* - replace each value x of ia with ib->val[x] */
    int n, x;
    int *val, *ibval, len; /* pointer for code optimization */
    
#ifdef DEBUG
    assert(ia != NULL);
    assert(ib != NULL);
#endif
    ibval = ib->val;
    len   = ia->len; 
    n = 0;
    val = ia->val;
    while (n < len) {
	x = *val;
	*val = ibval[x];
	val++; 
	n++;
    }
    /* ia-len is not changed */
}

void ia_map_slow(Intarray ia, Intarray ib)
{   /* - replace each value x of ia with ib->val[x] */
    int n, x;
    
#ifdef DEBUG
    assert(ia != NULL);
    assert(ib != NULL);
#endif
    for (n = 0; n < ia->len; n++) {
	x = ia->val[n];
	ia->val[n] = ib->val[x];
    }
    /* ia-len is not changed */
}

/*
 * Interface to Intarray package
 */


void ia_count(Intarray ia, Intarray count, int num_labels)
{
    int n;
    int *val, *tail, *iaval; 
    
#ifdef DEBUG
    assert(ia != NULL);
    assert(count != NULL);
#endif
    /* fill count with num_labels zeros */
    ia_realloc(count, num_labels);
    count->len = num_labels; 
    val = count->val;
    tail = val + num_labels; 
    while (val < tail)
        *(val++) = 0; /* count is zero */
    val = count->val;
    iaval = ia->val; 
    for (n = 0; n < ia->len; n++) {
	val[*(iaval++)]++;
  }
}

void ia_count_slow(Intarray ia, Intarray count, int num_labels)
{
    int n;
    
#ifdef DEBUG
    assert(ia != NULL);
    assert(count != NULL);
#endif
    ia_make_null(count); 
    for (n = 0; n < num_labels; n++)
      ia_push_back(count, 0);  /* count is zero */
    for (n = 0; n < ia->len; n++) {
	assert(ia->val[n] < count->len);
      count->val[ia->val[n]]++;
  }
}

void ia_scalar_add(Intarray ia, int x)
{
    int n;

#ifdef DEBUG
    assert(ia != NULL);
#endif
    for (n = 0; n < ia->len; n++) {
	ia->val[n] = ia->val[n] + x;
    }
}

/* no need for inline */
void ia_fill(Intarray ia, int len, int x)
{   /* re-initialize and fill the contiguous
     * len cells of ia with integer x */
    int n;
    int *val, *tail;

#ifdef DEBUG
    assert(ia != NULL);
#endif
    ia_realloc(ia, len);
    val = ia->val;
    tail = val + len; 
    while (val < tail)
        *(val++) = x; /* count is zero */
    ia->len = len; 
}

void ia_fill_org(Intarray ia, int len, int x)
{   /* re-initialize and fill the contiguous
     * len cells of ia with integer x */
    int n;

#ifdef DEBUG
    assert(ia != NULL);
#endif
    ia_realloc(ia, len);
    for (n = 0; n < len; n++) {
	ia->val[n] = x; 
    }
}


int ia_is_equal(Intarray ia, Intarray ib)
{ /* checks if ia and ib are same intarrays
   */
  int n;

#ifdef DEBUG
  assert(ia != NULL && ib != NULL); 
  assert(ia->maxlen >= ib->len);
#endif
  if (ia->len != ib->len)
    return 0;
  for (n = 0; n < ia->len; n++)
    if (ia->val[n] != ib->val[n])
      return 0;
  return 1;
}

/*
 * Generating test sequences for simulations. 
 */

void ia_fill_inc(Intarray ia, int len)
{   /* fill ia with the increasing integers 0, 1, ..., len-1.
     */
  int n;
  
  ia_realloc(ia, len);
  for (n = 0; n < len; n++) {
      ia->val[n] = n; 
  }
  ia->len = n;
}

void ia_fill_random(Intarray ia, int range, int len)
{   /* fill ia with len random elements whose values take
       the range from 0 to range */
    int n; 

    ia_make_null(ia);
    for (n = 0; n < len; n++) {
	ia_push_back(ia, (rand() % range)); 
    }
    ia->len = n;
}



/******************************************************************
 * End
 * 
 ******************************************************************/
